"""Unit tests for the GitLab CI integration.

Version Added:
    5.0
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, cast

import gitlab
import kgb
from django.urls import reverse
from djblets.conditions import ConditionSet, Condition
from gitlab.v4.objects import (
    Project,
    ProjectManager,
    ProjectPipeline,
    ProjectPipelineManager,
)
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.models import StatusUpdate
from reviewboard.reviews.signals import status_update_request_run

from rbintegrations.gitlabci.forms import GitLabCIIntegrationConfigForm
from rbintegrations.gitlabci.integration import GitLabCIIntegration
from rbintegrations.testing.testcases import IntegrationTestCase

if TYPE_CHECKING:
    from typing import Any

    from reviewboard.reviews.models import ReviewRequest


class BaseGitLabCITestCase(IntegrationTestCase[GitLabCIIntegration]):
    """Base class for GitLab CI tests.

    Version Added:
        5.0
    """

    integration_cls = GitLabCIIntegration
    fixtures = ['test_scmtools', 'test_site', 'test_users']

    def _create_config(
        self,
        *,
        self_hosted: bool = False,
        with_local_site: bool = False,
        run_manually: bool = False,
        use_trigger_token: bool = False,
        gitlab_name: (str | None) = None,
        gitlab_ref: (str | None) = None,
        gitlab_vars: (str | None) = None,
        gitlab_inputs: (str | None) = None,
        gitlab_webhook_secret_token: (str | None) = None,
    ) -> IntegrationConfig:
        """Create an integration config.

        Args:
            self_hosted (bool, optional):
                Whether to test with a self-hosted GitLab URL.

            with_local_site (bool, optional):
                Whether to test with a local site.

            run_manually (bool, optional):
                Whether to set up the integration configuration to run
                only when manually triggered.

            use_trigger_token (bool, optional):
                Whether to use trigger token instead of private token.

            gitlab_name (str, optional):
                Custom GitLab project name.

            gitlab_ref (str, optional):
                Custom GitLab ref.

            gitlab_vars (str, optional):
                Custom GitLab variables as JSON string.

            gitlab_inputs (str, optional):
                Custom GitLab inputs as JSON string.

            gitlab_webhook_secret_token (str, optional):
                Secret token for webhook validation.

        Returns:
            reviewboard.integrations.models.IntegrationConfig:
            The new integration configuration.
        """
        choice = ReviewRequestRepositoriesChoice()
        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator('any')),
        ])

        if with_local_site:
            self.local_site = self.get_local_site(name=self.local_site_name)
        else:
            self.local_site = None

        config = self.integration.create_config(
            name='Config 1',
            enabled=True,
            local_site=self.local_site,
            save=False,
        )
        config.set('conditions', condition_set.serialize())
        config.set('run_manually', run_manually)

        if self_hosted:
            config.set('gitlab_endpoint', 'https://gitlab.example.com')
        else:
            config.set('gitlab_endpoint', 'https://gitlab.com')

        if use_trigger_token:
            config.set('gitlab_token_type', 'trigger_token')
            config.set('gitlab_token', 'test-trigger-token')
        else:
            config.set('gitlab_token_type', 'private_token')
            config.set('gitlab_token', 'test-private-token')

        config.set('gitlab_name', gitlab_name or '{repository_name}')
        config.set('gitlab_ref', gitlab_ref or '{branch}')
        config.set('gitlab_vars', gitlab_vars or '{}')
        config.set('gitlab_inputs', gitlab_inputs or '{}')
        config.set('gitlab_report_job_state', False)

        if gitlab_webhook_secret_token:
            config.set('gitlab_webhook_secret_token',
                       gitlab_webhook_secret_token)

        config.save()

        return cast(IntegrationConfig, config)


class GitLabCIIntegrationTests(BaseGitLabCITestCase):
    """Unit tests for GitLab CI.

    Version Added:
        5.0
    """

    def test_build_new_review_request(self) -> None:
        """Testing GitLabCIIntegration builds a new review request"""
        review_request = self._setup_build_test()

        @self.spy_for(ProjectManager.get, owner=ProjectManager)
        def _get_project(*args, **kwargs) -> Project:
            return Project(args[0], {})

        @self.spy_for(ProjectPipelineManager.create,
                      owner=ProjectPipelineManager)
        def _create_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {})

        self.spy_on(gitlab.Gitlab.auth,
                    owner=gitlab.Gitlab,
                    call_original=False)

        review_request.publish(review_request.submitter)

        api_token = self.integration.get_or_create_api_token(
            user=self.integration.get_or_create_user(),
            local_site=self.local_site)

        self.assertSpyCalledWith(
            ProjectPipelineManager.create,
            data={
                'inputs': {},
                'ref': 'my-branch',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_API_TOKEN',
                        'value': api_token.token,
                    },
                    {
                        'key': 'REVIEWBOARD_DIFF_REVISION',
                        'value': 1,
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': self.config.pk,
                    },
                    {
                        'key': 'REVIEWBOARD_PIPELINE_NAME',
                        'value': 'Review Request #1: Test Summary',
                    },
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': review_request.display_id,
                    },
                    {
                        'key': 'REVIEWBOARD_SERVER',
                        'value': 'http://example.com/',
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': 1,
                    },
                ],
            })

    def test_build_new_review_request_with_trigger_token(self) -> None:
        """Testing GitLabCIIntegration builds a new review request with
        trigger token
        """
        review_request = self._setup_build_test(use_trigger_token=True)

        @self.spy_for(ProjectManager.get, owner=ProjectManager)
        def _get_project(*args, **kwargs) -> Project:
            return Project(args[0], {})

        @self.spy_for(Project.trigger_pipeline, owner=Project)
        def _trigger_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {})

        review_request.publish(review_request.submitter)

        api_token = self.integration.get_or_create_api_token(
            user=self.integration.get_or_create_user(),
            local_site=self.local_site)

        self.assertSpyCalledWith(
            Project.trigger_pipeline,
            'my-branch',
            'test-trigger-token',
            inputs={},
            variables={
                'REVIEWBOARD_API_TOKEN': api_token.token,
                'REVIEWBOARD_DIFF_REVISION': 1,
                'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID': self.config.pk,
                'REVIEWBOARD_PIPELINE_NAME': 'Review Request #1: Test Summary',
                'REVIEWBOARD_REVIEW_REQUEST': review_request.display_id,
                'REVIEWBOARD_SERVER': 'http://example.com/',
                'REVIEWBOARD_STATUS_UPDATE_ID': 1,
            })

    def test_build_new_review_request_with_local_site(self) -> None:
        """Testing GitLabCIIntegration builds a new review request with a
        local site
        """
        review_request = self._setup_build_test(with_local_site=True)

        @self.spy_for(ProjectManager.get, owner=ProjectManager)
        def _get_project(*args, **kwargs) -> Project:
            return Project(args[0], {'id': 1})

        @self.spy_for(ProjectPipelineManager.create,
                      owner=ProjectPipelineManager)
        def _create_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {'id': 1})

        self.spy_on(gitlab.Gitlab.auth,
                    owner=gitlab.Gitlab,
                    call_original=False)

        review_request.publish(review_request.submitter)

        api_token = self.integration.get_or_create_api_token(
            user=self.integration.get_or_create_user(),
            local_site=self.local_site)

        self.assertSpyCalledWith(
            ProjectPipelineManager.create,
            data={
                'inputs': {},
                'ref': 'my-branch',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_API_TOKEN',
                        'value': api_token.token,
                    },
                    {
                        'key': 'REVIEWBOARD_DIFF_REVISION',
                        'value': 1,
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': self.config.pk,
                    },
                    {
                        'key': 'REVIEWBOARD_PIPELINE_NAME',
                        'value': (
                            f'Review Request #{review_request.display_id}: '
                            f'Test Summary'
                        ),
                    },
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': review_request.display_id,
                    },
                    {
                        'key': 'REVIEWBOARD_SERVER',
                        'value': (
                            f'http://example.com/s/{self.local_site_name}/'
                        ),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': 1,
                    },
                ],
            })

    def test_manual_run_no_build_on_publish(self) -> None:
        """Testing GitLabCIIntegration with manual trigger does not
        build on publish
        """
        review_request = self._setup_build_test(run_manually=True)

        @self.spy_for(ProjectManager.get, owner=ProjectManager)
        def _get_project(*args, **kwargs) -> Project:
            return Project(args[0], {})

        @self.spy_for(ProjectPipelineManager.create,
                      owner=ProjectPipelineManager)
        def _create_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {})

        @self.spy_for(Project.trigger_pipeline, owner=Project)
        def _trigger_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {})

        review_request.publish(review_request.submitter)

        self.assertSpyNotCalled(ProjectPipelineManager.create)
        self.assertSpyNotCalled(Project.trigger_pipeline)

    def test_build_manual_run(self) -> None:
        """Testing GitLabCIIntegration build via a manual trigger"""
        review_request = self._setup_build_test(run_manually=True)

        @self.spy_for(ProjectManager.get, owner=ProjectManager)
        def _get_project(*args, **kwargs) -> Project:
            return Project(args[0], {})

        @self.spy_for(ProjectPipelineManager.create,
                      owner=ProjectPipelineManager)
        def _create_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {})

        self.spy_on(gitlab.Gitlab.auth,
                    owner=gitlab.Gitlab,
                    call_original=False)

        status_update = self.create_status_update(
            service_id='gitlab-ci',
            review_request=review_request)

        status_update_request_run.send(
            sender=self.__class__,
            status_update=status_update)

        self.assertSpyCalled(ProjectPipelineManager.create)

    def test_variable_replacement(self) -> None:
        """Testing GitLabCIIntegration variable replacement in project name
        and ref
        """
        review_request = self._setup_build_test(
            gitlab_name='{repository_name}-ci',
            gitlab_ref='{branch}',
            repository_name='Test Repo')

        @self.spy_for(ProjectManager.get, owner=ProjectManager)
        def _get_project(*args, **kwargs) -> Project:
            return Project(args[0], {})

        @self.spy_for(ProjectPipelineManager.create,
                      owner=ProjectPipelineManager)
        def _create_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {})

        self.spy_on(gitlab.Gitlab.auth,
                    owner=gitlab.Gitlab,
                    call_original=False)

        review_request.publish(review_request.submitter)

        self.assertSpyCalledWith(ProjectManager.get, 'Test Repo-ci')

    def test_variable_replacement_in_json_fields(self) -> None:
        """Testing GitLabCIIntegration variable replacement in JSON fields"""
        review_request = self._setup_build_test(
            gitlab_vars='{"REPO": "{repository_name}", "BR": "{branch}"}',
            gitlab_inputs='{"input1": "{repository_name}"}')

        @self.spy_for(ProjectManager.get, owner=ProjectManager)
        def _get_project(*args, **kwargs) -> Project:
            return Project(args[0], {})

        @self.spy_for(ProjectPipelineManager.create,
                      owner=ProjectPipelineManager)
        def _create_pipeline(*args, **kwargs) -> ProjectPipeline:
            return ProjectPipeline(args[0], {})

        self.spy_on(gitlab.Gitlab.auth,
                    owner=gitlab.Gitlab,
                    call_original=False)

        review_request.publish(review_request.submitter)

        api_token = self.integration.get_or_create_api_token(
            user=self.integration.get_or_create_user(),
            local_site=self.local_site)

        self.assertSpyCalledWith(
            ProjectPipelineManager.create,
            data={
                'inputs': {
                    'input1': 'Test Repo',
                },
                'ref': 'my-branch',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_API_TOKEN',
                        'value': api_token.token,
                    },
                    {
                        'key': 'REVIEWBOARD_DIFF_REVISION',
                        'value': 1,
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': self.config.pk,
                    },
                    {
                        'key': 'REVIEWBOARD_PIPELINE_NAME',
                        'value': 'Review Request #1: Test Summary',
                    },
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': review_request.display_id,
                    },
                    {
                        'key': 'REVIEWBOARD_SERVER',
                        'value': 'http://example.com/',
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': 1,
                    },
                    {
                        'key': 'REPO',
                        'value': 'Test Repo',
                    },
                    {
                        'key': 'BR',
                        'value': 'my-branch',
                    },
                ],
            })

    def _setup_build_test(
        self,
        *,
        with_local_site: bool = False,
        run_manually: bool = False,
        use_trigger_token: bool = False,
        gitlab_name: str | None = None,
        gitlab_ref: str | None = None,
        gitlab_vars: str | None = None,
        gitlab_inputs: str | None = None,
        repository_name: str = 'Test Repo',
    ) -> ReviewRequest:
        """Set up a basic build test.

        Args:
            with_local_site (bool, optional):
                Whether to create with a local site.

            run_manually (bool, optional):
                Whether to configure manual runs.

            use_trigger_token (bool, optional):
                Whether to use trigger token.

            gitlab_name (str, optional):
                Custom GitLab project name.

            gitlab_ref (str, optional):
                Custom GitLab ref.

            gitlab_vars (str, optional):
                Custom GitLab variables as JSON string.

            gitlab_inputs (str, optional):
                Custom GitLab inputs as JSON string.

            repository_name (str, optional):
                Name for the repository.

        Returns:
            reviewboard.reviews.models.ReviewRequest:
            The created review request.
        """
        repository = self.create_repository(
            name=repository_name,
            with_local_site=with_local_site)
        review_request = self.create_review_request(
            repository=repository,
            with_local_site=with_local_site,
            branch='my-branch')
        diffset = self.create_diffset(
            review_request=review_request,
            base_commit_id='8fd69d70f07b57c21ad8733c1c04ae604d21493f')
        diffset.save()

        self.config = self._create_config(
            with_local_site=with_local_site,
            run_manually=run_manually,
            use_trigger_token=use_trigger_token,
            gitlab_name=gitlab_name,
            gitlab_ref=gitlab_ref,
            gitlab_vars=gitlab_vars,
            gitlab_inputs=gitlab_inputs)
        self.integration.enable_integration()

        return review_request


class GitLabCIIntegrationConfigFormTests(BaseGitLabCITestCase):
    """Unit tests for GitLabCIIntegrationConfigForm.

    Version Added:
        5.0
    """

    def test_gitlabci_config_form_valid(self) -> None:
        """Testing GitLabCIIntegrationConfigForm validation success"""
        form = GitLabCIIntegrationConfigForm(
            integration=self.integration,
            request=None,
            data={
                'conditions_last_id': 0,
                'conditions_mode': 'always',
                'name': 'test',
                'gitlab_endpoint': 'https://gitlab.com',
                'gitlab_token_type': 'private_token',
                'gitlab_token': 'test-token',
                'gitlab_name': 'project-name',
                'gitlab_ref': 'main',
                'gitlab_inputs': '{}',
                'gitlab_vars': '{}',
            })

        self.spy_on(gitlab.Gitlab.auth,
                    owner=gitlab.Gitlab,
                    call_original=False)
        self.assertTrue(form.is_valid())

    def test_gitlabci_config_form_json_failure(self) -> None:
        """Testing GitLabCIIntegrationConfigForm validation failure with a JSON
        parse error
        """
        form = GitLabCIIntegrationConfigForm(
            integration=self.integration,
            request=None,
            data={
                'conditions_last_id': 0,
                'conditions_mode': 'always',
                'name': 'test',
                'gitlab_endpoint': 'https://gitlab.com',
                'gitlab_token_type': 'private_token',
                'gitlab_token': 'test-token',
                'gitlab_name': 'project-name',
                'gitlab_ref': 'main',
                'gitlab_inputs': 'invalid json',
                'gitlab_vars': '{}',
            })

        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors['gitlab_inputs'],
            [
                'Unable to parse JSON: Expecting value: line 1 column 1 '
                '(char 0)',
            ])

    def test_gitlabci_config_form_auth_failure(self) -> None:
        """Testing GitLabCIIntegrationConfigForm validation failure with
        authentication failure
        """
        form = GitLabCIIntegrationConfigForm(
            integration=self.integration,
            request=None,
            data={
                'conditions_last_id': 0,
                'conditions_mode': 'always',
                'name': 'test',
                'gitlab_endpoint': 'https://gitlab.com',
                'gitlab_token_type': 'private_token',
                'gitlab_token': 'bad-token',
                'gitlab_name': 'project-name',
                'gitlab_ref': 'main',
                'gitlab_inputs': '{}',
                'gitlab_vars': '{}',
            })

        self.spy_on(
            gitlab.Gitlab.auth,
            owner=gitlab.Gitlab,
            op=kgb.SpyOpRaise(
                gitlab.GitlabAuthenticationError('Authentication failed.')
            ))
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['gitlab_endpoint'],
                         ['Authentication failed.'])


class GitLabCIWebHookTests(BaseGitLabCITestCase):
    """Unit tests for the GitLab CI webhook handler.

    Version Added:
        5.0
    """

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        self.repository = self.create_repository()
        self.review_request = self.create_review_request(
            repository=self.repository,
            public=True)

        self.status_update = self.create_status_update(self.review_request)

        self.config = self._create_config()
        self.integration.enable_integration()

        self.webhook_url = reverse('gitlab-ci-webhook')

    def test_webhook_bad_json(self) -> None:
        """Testing GitLabCIWebHookView with invalid JSON"""
        response = self.client.post(
            self.webhook_url,
            data='invalid json',
            content_type='application/json')

        self.assertEqual(response.status_code, 400)

    def test_webhook_missing_object_kind(self) -> None:
        """Testing GitLabCIWebHookView with missing object_kind"""
        payload = json.dumps({})
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, b'Missing pipeline event')

    def test_webhook_wrong_object_kind(self) -> None:
        """Testing GitLabCIWebHookView with wrong object_kind"""
        payload = json.dumps({'object_kind': 'build'})
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, b'Missing pipeline event')

    def test_webhook_missing_variables(self) -> None:
        """Testing GitLabCIWebHookView with missing variables"""
        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'success',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': None,
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 202)

    def test_webhook_missing_required_variables(self) -> None:
        """Testing GitLabCIWebHookView with missing required variables"""
        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'success',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {'key': 'OTHER_VAR', 'value': 'value'},
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 202)

    def test_webhook_invalid_config_id(self) -> None:
        """Testing GitLabCIWebHookView with invalid config ID"""
        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'success',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': str(self.status_update.pk),
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': '99999',
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 400)
        self.assertIn(b'Unable to find integration config', response.content)

    def test_webhook_wrong_secret_token(self) -> None:
        """Testing GitLabCIWebHookView with wrong secret token"""
        self.config.set('gitlab_webhook_secret_token', 'correct-secret')
        self.config.save()

        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'success',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': str(self.status_update.pk),
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': str(self.config.pk),
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json',
            HTTP_X_GITLAB_TOKEN='wrong-secret')

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.content,
                         b'WebHook secret token does not match')

    def test_webhook_invalid_status_update_id(self) -> None:
        """Testing GitLabCIWebHookView with invalid status update ID"""
        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'success',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': '99999',
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': str(self.config.pk),
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 400)
        self.assertIn(b'Unable to find matching status update',
                      response.content)

    def test_webhook_pipeline_success(self) -> None:
        """Testing GitLabCIWebHookView with successful pipeline"""
        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [
                {
                    'id': 123,
                    'name': 'test-job',
                    'stage': 'test',
                    'status': 'success',
                },
            ],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'success',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': str(self.status_update.pk),
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': str(self.config.pk),
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 200)

        status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        self.assertEqual(status_update.state, StatusUpdate.DONE_SUCCESS)
        self.assertEqual(status_update.url,
                         'https://gitlab.example.com/pipeline/123')

    def test_webhook_pipeline_failure(self) -> None:
        """Testing GitLabCIWebHookView with failed pipeline"""
        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'failed',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': str(self.status_update.pk),
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': str(self.config.pk),
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 200)

        status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        self.assertEqual(status_update.state, StatusUpdate.ERROR)

    def test_webhook_pipeline_running(self) -> None:
        """Testing GitLabCIWebHookView with running pipeline"""
        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'running',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': str(self.status_update.pk),
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': str(self.config.pk),
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 200)

        status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        self.assertEqual(status_update.state, StatusUpdate.PENDING)

    def test_webhook_with_job_reporting(self) -> None:
        """Testing GitLabCIWebHookView with job state reporting enabled"""
        self.config.set('gitlab_report_job_state', True)
        self.config.save()

        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [
                {
                    'id': 123,
                    'name': 'test-job-1',
                    'stage': 'test',
                    'status': 'success',
                },
                {
                    'id': 124,
                    'name': 'test-job-2',
                    'stage': 'test',
                    'status': 'failed',
                },
            ],
            'object_attributes': {
                'id': 123,
                'source': 'api',
                'stages': ['test'],
                'status': 'failed',
                'url': 'https://gitlab.example.com/pipeline/123',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': str(self.status_update.pk),
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': str(self.config.pk),
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 200)

        status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        self.assertEqual(
            status_update.extra_data.get('gitlab_ci_builds', {}),
            {
                '123': {
                    'name': 'test-job-1',
                    'stage': 'test',
                    'status': 'success',
                    'url': 'https://gitlab.example.com/project/-/jobs/123',
                },
                '124': {
                    'name': 'test-job-2',
                    'stage': 'test',
                    'status': 'failed',
                    'url': 'https://gitlab.example.com/project/-/jobs/124',
                },
            })

    def test_webhook_child_pipeline(self) -> None:
        """Testing GitLabCIWebHookView with child pipeline event"""
        self.config.set('gitlab_report_job_state', True)
        self.config.save()

        payload = json.dumps({
            'object_kind': 'pipeline',
            'builds': [
                {
                    'id': 125,
                    'name': 'child-job',
                    'stage': 'test',
                    'status': 'success',
                },
            ],
            'object_attributes': {
                'id': 124,
                'source': 'parent_pipeline',
                'stages': ['test'],
                'status': 'success',
                'url': 'https://gitlab.example.com/pipeline/124',
                'variables': [
                    {
                        'key': 'REVIEWBOARD_REVIEW_REQUEST',
                        'value': str(self.review_request.display_id),
                    },
                    {
                        'key': 'REVIEWBOARD_STATUS_UPDATE_ID',
                        'value': str(self.status_update.pk),
                    },
                    {
                        'key': 'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID',
                        'value': str(self.config.pk),
                    },
                ],
            },
            'project': {
                'web_url': 'https://gitlab.example.com/project',
            },
        })
        response = self.client.post(
            self.webhook_url,
            data=payload,
            content_type='application/json')

        self.assertEqual(response.status_code, 200)

        status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        gitlab_ci_builds = status_update.extra_data.get('gitlab_ci_builds', {})
        self.assertIn('125', gitlab_ci_builds)



"""Views for the GitLab CI integration.

Version Added:
    5.0
"""

from __future__ import annotations

import json
import logging
from typing import List, Literal, Optional, TYPE_CHECKING

import pydantic
from django.core.exceptions import ObjectDoesNotExist
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseServerError,
)
from django.utils.translation import gettext as _
from django.views.generic import View
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.models import GeneralComment, Review, StatusUpdate

from rbintegrations.gitlabci.integration import (
    GitLabCIIntegration,
    GitLabCIConfiguration,
)

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any, ClassVar

    from django.http import HttpRequest
    from djblets.util.typing import JSONValue
    from typing_extensions import TypeAlias


logger = logging.getLogger(__name__)


#: The possible status values for pipelines and jobs.
#:
#: This is retrieved from
#: https://gitlab.com/gitlab-org/gitlab/-/blob/master/app/models/concerns/ci/has_status.rb
#:
#: This may or may not be an authoritative list of the possible status values
#: but it's the best I've found.
#:
#: Version Added:
#:     5.0
GitLabStatus: TypeAlias = Literal[
    'canceled',
    'canceling',
    'created',
    'failed',
    'manual',
    'pending',
    'preparing',
    'running',
    'scheduled',
    'skipped',
    'success',
    'waiting_for_callback',
    'waiting_for_resource',
]


class _WebHookPayload(pydantic.BaseModel):
    """Data for the payload from the GitLab WebHook pipeline event.

    Version Added:
        5.0
    """

    class BuildData(pydantic.BaseModel):
        id: int
        name: str
        stage: str
        status: GitLabStatus

    class ObjectAttributes(pydantic.BaseModel):
        class PipelineVariable(pydantic.BaseModel):
            key: str
            value: str

        id: int
        source: str
        stages: List[str]
        status: GitLabStatus
        url: str
        variables: Optional[List[PipelineVariable]] = None

    class ProjectData(pydantic.BaseModel):
        web_url: str

    builds: List[BuildData]
    object_attributes: ObjectAttributes
    project: ProjectData


class GitlabCIWebHookView(View):
    """A view to handle webhook notifications from a GitLab CI build.

    Version Added:
        5.0
    """

    STATUS_STATE_MAP: ClassVar[Mapping[GitLabStatus, str]] = {
        'canceled': StatusUpdate.ERROR,  # Handled specially below.
        'canceling': StatusUpdate.PENDING,
        'created': StatusUpdate.PENDING,
        'failed': StatusUpdate.ERROR,
        'manual': StatusUpdate.NOT_YET_RUN,
        'pending': StatusUpdate.PENDING,
        'preparing': StatusUpdate.PENDING,
        'running': StatusUpdate.PENDING,
        'scheduled': StatusUpdate.PENDING,  # Handled specially below.
        'skipped': StatusUpdate.ERROR,
        'success': StatusUpdate.DONE_SUCCESS,
        'waiting_for_callback': StatusUpdate.PENDING,
        'waiting_for_resource': StatusUpdate.PENDING,
    }

    PIPELINE_STATUS_DESCRIPTION_MAP: ClassVar[Mapping[GitLabStatus, str]] = {
        'canceled': 'pipeline was canceled before completion.',
        'canceling': 'pipeline is in the process of cancelling.',
        'created': 'pipeline has been created.',
        'failed': 'pipeline failed.',
        'manual': 'pipeline needs to be manually run.',
        'pending': 'pipeline has not started running yet.',
        'preparing': 'pipeline is preparing to run.',
        'running': 'pipeline is running.',
        'scheduled': 'pipeline is scheduled to run.',
        'skipped': 'pipeline was skipped.',
        'success': 'pipeline completed successfully.',
        'waiting_for_callback': 'pipeline is waiting for an external action.',
        'waiting_for_resource': (
            'pipeline is waiting for a resource such as a runner.'
        ),
    }

    JOB_STATUS_DESCRIPTION_MAP: ClassVar[Mapping[GitLabStatus, str]] = {
        'canceled': 'cancelled',
        'canceling': 'cancelling',
        'created': 'created',
        'failed': 'failed',
        'manual': 'needs to be manually run',
        'pending': 'not started running yet',
        'preparing': 'preparing to run',
        'running': 'running',
        'scheduled': 'scheduled',
        'skipped': 'skipped',
        'success': 'succeeded',
        'waiting_for_callback': 'waiting for an external action',
        'waiting_for_resource': 'waiting for a resource',
    }

    def post(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle the POST.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple):
                Additional positional arguments, parsed from the URL.

            **kwargs (dict):
                Additional keyword arguments, parsed from the URL.

        Returns:
            django.http.HttpResponse:
            A response.
        """
        try:
            decoded: JSONValue = json.loads(request.body)
        except json.JSONDecodeError as e:
            return HttpResponseBadRequest(str(e))

        if (not isinstance(decoded, dict) or
            decoded.get('object_kind') != 'pipeline'):
            return HttpResponseBadRequest('Missing pipeline event')

        try:
            payload = _WebHookPayload.model_validate(decoded)
        except pydantic.ValidationError as e:
            error = f'Unable to validate WebHook payload: {e}'
            logger.error('gitlab-ci webhook: %s',
                         error,
                         extra={'request': request})
            return HttpResponseBadRequest(error)

        object_attributes = payload.object_attributes

        if object_attributes.variables is None:
            # This WebHook notification was for a pipeline that was not
            # triggered by Review Board. Report a 202 code back to GitLab.
            return HttpResponse(status=202)

        variables = {
            item.key: item.value
            for item in object_attributes.variables
        }

        if ('REVIEWBOARD_REVIEW_REQUEST' not in variables or
            'REVIEWBOARD_STATUS_UPDATE_ID' not in variables or
            'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID' not in variables):
            # This WebHook notification was for a pipeline that was not
            # triggered by Review Board. Report a 202 code back to GitLab.
            return HttpResponse(status=202)

        pipeline_id = object_attributes.id

        try:
            review_request_id = int(variables['REVIEWBOARD_REVIEW_REQUEST'])
            status_update_id = int(variables['REVIEWBOARD_STATUS_UPDATE_ID'])
            config_id = int(
                variables['REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID'])
        except ValueError:
            error = 'Variables were not in expected format.'
            logger.error('gitlab-ci webhook (pipeline %s): %s',
                         pipeline_id, error,
                         extra={'request': request})

            return HttpResponseBadRequest(error)

        logger.debug(
            'gitlab-ci webhook (pipeline %s): processing webhook for review '
            'request %s (status update %s)',
            pipeline_id,
            review_request_id,
            status_update_id,
            extra={'request': request},
        )

        try:
            integration_config = IntegrationConfig.objects.get(pk=config_id)
        except ObjectDoesNotExist:
            error = f'Unable to find integration config {config_id}'
            logger.error('gitlab-ci webhook (pipeline %s): %s',
                         pipeline_id, error,
                         extra={'request': request})

            return HttpResponseBadRequest(error)

        try:
            gitlab_config = GitLabCIConfiguration.model_validate(
                integration_config.settings)
        except pydantic.ValidationError as e:
            error = (
                f'Error validating GitLab CI configuration for '
                f'integration config pk={integration_config.pk}: {e}'
            )
            logger.error('gitlab-ci webhook (pipeline %s): %s',
                         pipeline_id, error,
                         extra={'request': request})

            return HttpResponseServerError(error)

        webhook_token = request.META.get('HTTP_X_GITLAB_TOKEN', '')
        expected_webhook_token = gitlab_config.gitlab_webhook_secret_token

        if (expected_webhook_token is not None and
            webhook_token != expected_webhook_token):
            error = 'WebHook secret token does not match'
            logger.error('gitlab-ci webhook (pipeline %s): %s',
                         pipeline_id, error,
                         extra={'request': request})

            return HttpResponseForbidden(error)

        try:
            status_update = StatusUpdate.objects.get(pk=status_update_id)
        except ObjectDoesNotExist:
            error = \
                f'Unable to find matching status update ID {status_update_id}'
            logger.error('gitlab-ci webhook (pipeline %s): %s',
                         pipeline_id, error,
                         extra={'request': request})

            return HttpResponseBadRequest(error)

        if status_update.review_request.display_id != review_request_id:
            error = (
                f'Wrong review request {review_request_id} for status update '
                f'ID {status_update_id}'
            )
            logger.error('gitlab-ci webhook (pipeline %s): %s',
                         pipeline_id, error,
                         extra={'request': request})

            return HttpResponseBadRequest(error)

        integration = integration_config.integration
        assert isinstance(integration, GitLabCIIntegration)

        try:
            self._update_from_payload(
                status_update=status_update,
                gitlab_config=gitlab_config,
                integration=integration,
                pipeline_id=pipeline_id,
                payload=payload)
        except Exception as e:
            logger.error('gitlab-ci webhook (pipeline %s): %s',
                         pipeline_id, e,
                         extra={'request': request})

            return HttpResponseBadRequest(e)

        return HttpResponse()

    def _update_from_payload(
        self,
        *,
        status_update: StatusUpdate,
        gitlab_config: GitLabCIConfiguration,
        integration: GitLabCIIntegration,
        pipeline_id: int,
        payload: _WebHookPayload,
    ) -> None:
        """Update the status update with the data from the payload.

        Args:
            status_update (reviewboard.reviews.models.StatusUpdate):
                The status update.

            gitlab_config (rbintegrations.gitlabci.integration.
                           GitLabCIConfiguration):
                The configuration settings.

            integration (rbintegrations.gitlabci.integration.
                         GitLabCIIntegration):
                The integration instance.

            pipeline_id (int):
                The ID of the pipeline triggering the WebHook event.

            payload (_WebHookPayload):
                The payload from the WebHook.

        Raises:
            Exception:
                An error occurred while processing the payload.
        """
        object_attributes = payload.object_attributes
        source = object_attributes.source

        if source == 'api':
            # This is an update for the main pipeline that we ourselves
            # triggered.
            build_status = object_attributes.status

            try:
                # The CANCELLED state was added in Review Board 7.1. If it's
                # available, we want to use it. If not, fall back on the
                # definition in STATUS_STATE_MAP.
                if (build_status in {'canceled', 'skipped'} and
                    hasattr(StatusUpdate, 'CANCELLED')):
                    new_state = getattr(StatusUpdate, 'CANCELLED')
                else:
                    new_state = self.STATUS_STATE_MAP[build_status]
            except KeyError:
                raise Exception(f'Unknown build status {build_status}')

            url = object_attributes.url
            assert isinstance(url, str)

            if gitlab_config.gitlab_report_job_state:
                self._update_build_status(payload, status_update,
                                          integration)
                status_extra_fields = ['extra_data', 'review']
            else:
                status_extra_fields = None

            integration.update_status(
                status_update,
                state=new_state,
                description=self.PIPELINE_STATUS_DESCRIPTION_MAP[build_status],
                url=url,
                url_text=_('View pipeline'),
                extra_fields=status_extra_fields,
            )
        elif source == 'parent_pipeline':
            # This is an update for a child pipeline that was triggered from
            # the main pipeline. We don't want to change the state, URL, or
            # other fields, but we do want to update build status if
            # appropriate.
            if gitlab_config.gitlab_report_job_state:
                self._update_build_status(payload, status_update,
                                          integration)
                integration.update_status(
                    status_update,
                    extra_fields=['extra_data', 'review'])

    def _update_build_status(
        self,
        payload: _WebHookPayload,
        status_update: StatusUpdate,
        integration: GitLabCIIntegration,
    ) -> None:
        """Update the build status for the pipeline.

        Args:
            payload (_WebHookPayload):
                The webhook payload.

            status_update (reviewboard.reviews.models.StatusUpdate):
                The status update for the pipeline.

            integration (rbintegrations.gitlabci.integration.
                         GitLabCIIntegration):
                The integration instance.
        """
        build_info = status_update.extra_data.get('gitlab_ci_builds', {})
        pipeline_info = status_update.extra_data.get('gitlab_ci_pipelines', {})
        project_url = payload.project.web_url
        object_attributes = payload.object_attributes

        if object_attributes.source == 'parent_pipeline':
            # This event is from a child pipeline.
            pipeline_id = str(object_attributes.id)

            pipeline_info[pipeline_id] = {
                'stages': object_attributes.stages,
                'status': object_attributes.status,
                'url': object_attributes.url,
            }

        for build in payload.builds:
            job_id = str(build.id)

            build_info[job_id] = {
                'name': build.name,
                'stage': build.stage,
                'status': build.status,
                'url': f'{project_url}/-/jobs/{job_id}',
            }

        status_update.extra_data['gitlab_ci_builds'] = build_info
        status_update.extra_data['gitlab_ci_pipelines'] = pipeline_info

        job_status = self._make_status_text(pipeline_info, build_info)

        if status_update.review:
            review = status_update.review

            try:
                comment = review.general_comments.all()[0]
                comment.text = job_status
                comment.save(update_fields=['text'])
            except ObjectDoesNotExist:
                logger.error(
                    'Unable to get general comment for GitLab CI job '
                    'summary review pk=%s',
                    review.pk)
        else:
            # Ideally we'd publish a review with the job summary as the
            # body_top, but the _status_update_review_section.html template in
            # Review Board only includes review content if they have any
            # comments. We therefore stick the job summary into a general
            # comment.
            review = Review.objects.create(
                review_request=status_update.review_request,
                user=integration.get_or_create_user(),
                public=False,
            )
            comment = GeneralComment.objects.create(
                text=job_status,
                rich_text=True,
            )
            review.general_comments.add(comment)
            review.publish()

            status_update.review = review
            status_update.save(update_fields=['review'])

    def _make_status_text(
        self,
        pipeline_info: Mapping[str, Mapping[str, Any]],
        build_info: Mapping[str, Mapping[str, Any]],
    ) -> str:
        """Return the text to use for build status in the review.

        Args:
            pipeline_info (dict):
                The stored child pipeline info.

            build_info (dict):
                The stored build info.

        Returns:
            str:
            Text for the build status.
        """
        status_lines: list[str] = []

        if pipeline_info:
            status_lines.append('# Child pipelines:')

            # Keys are stored as strings but we want to sort them numerically.
            for pipeline_id in sorted(pipeline_info.keys(), key=int):
                info = pipeline_info[pipeline_id]
                status = info['status']
                url = info['url']

                try:
                    description = self.PIPELINE_STATUS_DESCRIPTION_MAP[status]
                except KeyError:
                    logger.error('gitlab-ci webhook: Unknown job status '
                                 'value %s for pipeline %s',
                                 status, pipeline_id)
                    description = 'unknown'

                status_lines.append(
                    f'[{pipeline_id}]({url}): {description}'
                )

        status_lines.append('# Job summary:')

        # Keys are stored as strings but we want to sort them numerically.
        for build_id in sorted(build_info.keys(), key=int):
            info = build_info[build_id]
            url = info['url']
            name = info['name']
            status = info['status']

            try:
                description = self.JOB_STATUS_DESCRIPTION_MAP[status]
            except KeyError:
                logger.error('gitlab-ci webhook: Unknown job status value %s '
                             'for job %s',
                             status, build_id)
                description = 'unknown'

            status_lines.append(
                f'[{name}]({url}): {description}'
            )

        return '\n'.join(status_lines)



"""Forms for Microsoft Teams Integration.

Version Added:
    4.0
"""

from __future__ import annotations

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm
from reviewboard.reviews.conditions import ReviewRequestConditionChoices

from rbintegrations.util.conditions import ReviewRequestConditionsField


MS_TEAMS_INTEGRATION_DOCS_URL = (
    'https://www.reviewboard.org/integrations/'
    'microsoft-teams/#microsoft-teams-setup'
)


class MSTeamsIntegrationConfigForm(IntegrationConfigForm):
    """Admin configuration form for Microsoft Teams.

    This allows an administrator to set up a chat configuration for sending
    messages to a given Microsoft Teams WebHook URL based on the specified
    conditions.

    Version Added:
        4.0
    """

    conditions = ReviewRequestConditionsField()

    webhook_url = forms.CharField(
        label=_('WebHook URL'),
        required=True,
        help_text=_('Your Microsoft Teams WebHook URL.'),
        widget=forms.TextInput(attrs={
            'size': 80,
        }))

    class Meta:
        fieldsets = (
            ('What To Post', {
                'description': _(
                    'You can choose which review requests would be posted by '
                    'choosing the repositories and groups to match against.'
                ),
                'fields': ('conditions',),
            }),
            ('Where To Post', {
                'description': _(
                    f'Create a new Incoming WebHook on Microsoft Teams then '
                    f'enter its WebHook URL below. See '
                    f'<a href="{MS_TEAMS_INTEGRATION_DOCS_URL}">'
                    f'here</a> for more information.'
                ),
                'fields': ('webhook_url',),
                'classes': ('wide',)
            }),
        )



"""Integration for associating review requests with Trello cards."""

from django.utils.functional import cached_property
from reviewboard.extensions.hooks import ReviewRequestFieldsHook
from reviewboard.integrations.base import Integration

from rbintegrations.trello.fields import TrelloField
from rbintegrations.trello.forms import TrelloIntegrationConfigForm


class TrelloIntegration(Integration):
    """Integrates Review Board with Trello."""

    name = 'Trello'
    description = 'Associate Trello cards with your review requests.'
    config_form_cls = TrelloIntegrationConfigForm

    def initialize(self):
        """Initialize the integration hooks."""
        ReviewRequestFieldsHook(self, 'main', [TrelloField])

    @cached_property
    def icon_static_urls(self):
        """The icons used for the integration."""
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        return {
            '1x': extension.get_static_url('images/trello/icon.png'),
            '2x': extension.get_static_url('images/trello/icon@2x.png'),
        }



"""Forms for Matrix chat integrations."""

from django import forms
from django.utils.translation import gettext_lazy as _
from reviewboard.integrations.forms import IntegrationConfigForm

from rbintegrations.util.conditions import ReviewRequestConditionsField


class MatrixIntegrationConfigForm(IntegrationConfigForm):
    """Form for configuring Matrix integration.

    This allows an administrator to set up a chat configuration for sending
    messages based on a given access token based on the specified
    conditions.
    """

    conditions = ReviewRequestConditionsField()

    access_token = forms.CharField(
        label=_('Access Token'),
        required=True,
        help_text=_('Your unique access Token. In Element, this can be found '
                    'by going to Profile > All settings > Help & About > '
                    'Advanced.'),
        widget=forms.TextInput(attrs={
            'size': 80,
        }))

    room_id = forms.CharField(
        label=_('Room ID'),
        required=True,
        help_text=_('The ID of the room that review request updates will be '
                    'sent to. In Element, this can be found by clicking on '
                    'the 3 dots beside the room icon to access Settings and '
                    'clicking on the "Advanced" tab.'),
        widget=forms.TextInput(attrs={
            'size': 80,
        }))

    server = forms.CharField(
        label=_('Server URL'),
        required=True,
        help_text=_('The server HTTP requests will be sent to (e.g., '
                    'https://matrix.org).'),
        widget=forms.TextInput(attrs={
            'size': 80,
        }))

    class Meta:
        fieldsets = (
            ('What To Post', {
                'description': _(
                    'You can choose which review requests would be posted by '
                    'choosing the repositories and groups to match against.'
                ),
                'fields': ('conditions',),
            }),
            ('Where To Post', {
                'description': _(
                    'To start, create a Matrix account on and add a new room. '
                    'For best compatibility, we recommend using element.io '
                    'for this.'

                ),
                'fields': ('access_token', 'room_id', 'server'),
                'classes': ('wide',),
            }),
        )



"""URL Request utilities."""

from urllib.request import Request as BaseURLRequest


class URLRequest(BaseURLRequest):
    """A request that can use any HTTP method.

    By default, the :py:class:`urllib2.Request` class only supports HTTP GET
    and HTTP POST methods. This subclass allows for any HTTP method to be
    specified for the request.
    """

    def __init__(self, url, body='', headers=None, method='GET'):
        """
        Initialize the URLRequest.

        Args:
            url (unicode):
                The URL to make the request against.

            body (unicode or bytes):
                The content of the request.

            headers (dict, optional):
                Additional headers to attach to the request.

            method (unicode, optional):
                The request method. If not provided, it defaults to a ``GET``
                request.
        """
        BaseURLRequest.__init__(self, url, body, headers or {})
        self.method = method

    def get_method(self):
        """Return the HTTP method of the request.

        Returns:
            unicode:
            The HTTP method of the request.
        """
        return self.method



"""URL definitions for the Trello integration."""

from django.urls import include, path

from rbintegrations.trello.views import TrelloCardSearchView


localsite_urlpatterns = [
    path('card-search/<int:review_request_id>/',
         TrelloCardSearchView.as_view(),
         name='trello-card-search'),
]

urlpatterns = [
    path('s/<local_site_name>/', include(localsite_urlpatterns)),
]
urlpatterns += localsite_urlpatterns



"""Compatibility logging code for modern versions of Review Board.

Version Added:
    4.0.2
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from djblets.log import (TimedLogInfo as DjbletsTimedLogInfo,
                         log_timed as djblets_log_timed)

if TYPE_CHECKING or hasattr(DjbletsTimedLogInfo, '__enter__'):
    TimedLogInfo = DjbletsTimedLogInfo
    log_timed = djblets_log_timed
else:
    import logging
    from datetime import datetime, timezone
    from typing import Literal, Mapping, Optional
    from uuid import uuid4

    if TYPE_CHECKING:
        from django.http import HttpRequest

    class TimedLogInfo(DjbletsTimedLogInfo):
        """Tracks the time between operations for logging purposes.

        This is created and returned by :py:func:`log_timed` to track how long
        an operation takes, warning or critical-erroring if it takes too long.

        This class is a backport compatibility class for
        :py:class:`djblets.log.TimedLogInfo`. It will be removed when
        Djblets 5.3 is the minimum version required for Integrations.

        Version Added:
            4.0.2
        """

        ######################
        # Instance variables #
        ######################

        #: Extra information to include with all log items.
        #:
        #: This will be populated with ``request``.
        #:
        #: Version Added:
        #:     5.3
        extra: Mapping[str, object]

        #: The logger used for all log entries.
        #:
        #: Version Added:
        #:     5.3
        logger: logging.Logger

        #: The trace ID to include in log messages.
        #:
        #: This is used to help follow the chain of events in a series of logs.
        #:
        #: Version Added:
        #:     5.3
        trace_id: str

        def __init__(
            self,
            *,
            message: str,
            warning_at: float,
            critical_at: float,
            default_level: int,
            log_beginning: bool,
            request: Optional[HttpRequest],
            extra: Mapping[str, object] = {},
            logger: Optional[logging.Logger] = None,
            trace_id: Optional[str] = None,
        ) -> None:
            """Initialize the state for the timer.

            Args:
                message (str):
                    The message to show for the log entries.,

                warning_at (float):
                    The number of seconds at which to log warnings.

                    This may contain fractions of seconds.

                critical_at (float):
                    The number of seconds at which to log critical errors.

                    This may contain fractions of seconds.

                default_level (int):
                    The default log level for the timing information.

                log_beginning (bool):
                    Whether to log the beginning time for the operation.

                request (django.http.HttpRequest, optional):
                    The optional HTTP request associated with the operation.

                extra (dict, optional):
                    Extra information to include with all log items.

                    This will be populated with ``request`` and ``trace_id``.

                logger (logging.Logger, optional):
                    The logger used for all log entries.

                trace_id (str, optional):
                    The trace ID to include in log messages.

                    This is used to help follow the chain of events in a
                    series of logs.
            """
            self.message = message
            self.warning_at = warning_at
            self.critical_at = critical_at
            self.default_level = default_level
            self.start_time = datetime.now(timezone.utc)
            self.request = request

            if logger is None:
                logger = logging.getLogger()

            if not trace_id:
                trace_id = str(uuid4())

            extra = {
                **extra,
                'request': self.request,
                'trace_id': trace_id,
            }

            self.extra = extra
            self.logger = logger
            self.trace_id = trace_id

            if log_beginning:
                logger.log(default_level,
                           f'[%s] Begin: {message}',
                           trace_id,
                           extra=extra)

        def __enter__(self) -> TimedLogInfo:
            """Enter the context for the timer.

            This will open the context and return itself, allowing the timer
            to automatically complete when the context ends.

            Context:
                TimedLogInfo:
                This object.
            """
            return self

        def __exit__(self, *args, **kwargs) -> Literal[False]:
            """Exit the context for the timer.

            This will log the end time of the operation, stopping the timer.

            Args:
                *args (tuple, unused):
                    Unused positional arguments.

                **kwargs (dict, unused):
                    Unused keyword arguments.
            """
            self.done()

            return False

        def done(self) -> None:
            """Stop the timed logging operation.

            The resulting time of the operation will be written to the log
            file.  The log level depends on how long the operation takes.
            """
            delta = datetime.now(timezone.utc) - self.start_time
            level = self.default_level
            logger = self.logger
            message = self.message
            extra = self.extra
            trace_id = self.trace_id
            total_seconds = delta.total_seconds()

            if total_seconds >= self.critical_at:
                level = logging.CRITICAL
            elif total_seconds >= self.warning_at:
                level = logging.WARNING

            logger.log(self.default_level,
                       f'[%s] End: {message}',
                       trace_id,
                       extra=extra)
            logger.log(level,
                       f'[%s] {message} took %d.%06d seconds',
                       trace_id,
                       delta.seconds,
                       delta.microseconds,
                       extra=extra)

    def log_timed(
        message: str,
        *,
        warning_at: float = 5,
        critical_at: float = 15,
        log_beginning: bool = True,
        default_level: int = logging.DEBUG,
        request: Optional[HttpRequest] = None,
        extra: Mapping[str, object] = {},
        logger: Optional[logging.Logger] = None,
        trace_id: Optional[str] = None,
    ) -> TimedLogInfo:
        """Times an operation, logging timing information.

        This will display a log message at the start of an operation and at the
        end, displaying the time taken for the operation. The final log entry's
        level will depend on the amount of time taken, switching to a warning
        if at ``warning_at`` seconds and a critical error at ``critical_at``
        seconds.

        This function can be called directly or used as a context manager.

        This class is a backport compatibility class for
        :py:class:`djblets.log.TimedLogInfo`. It will be removed when
        Djblets 5.3 is the minimum version required for Integrations.

        Version Added:
            4.0.2

        Args:
            message (str):
                The message to show for the log entries.,

            warning_at (int):
                The number of seconds at which to log warnings.

                This may contain fractions of seconds.

            critical_at (int):
                The number of seconds at which to log critical errors.

                This may contain fractions of seconds.

            default_level (int):
                The default log level for the timing information.

            log_beginning (bool):
                Whether to log the beginning time for the operation.

            request (django.http.HttpRequest, optional):
                The optional HTTP request associated with the operation.

            extra (dict, optional):
                Extra information to include with all log items.

                This will be populated with ``request`` and ``trace_id``.

            logger (logging.Logger, optional):
                The logger used for all log entries.

            trace_id (str, optional):
                The trace ID to include in log messages.

                This is used to help follow the chain of events in a series of
                logs.

                If not provided, one will be generated. This can then be
                accessed on :py:attr:`TimedLogInfo.trace_id`.

        Example:
            .. code-block:: python

               from djblets.log import log_timed

               # As a direct function call:
               t = log_timed('Doing a thing')

               try:
                   ...
               finally:
                   t.done()

               # As a context manager:
               with log_timed('Doing a thing') as t:
                   ...
        """
        return TimedLogInfo(message=message,
                            warning_at=warning_at,
                            critical_at=critical_at,
                            default_level=default_level,
                            log_beginning=log_beginning,
                            request=request,
                            extra=extra,
                            logger=logger)



"""RBIntegrations version and package information.

These variables and functions can be used to identify the version of
the module. They're largely used for packaging purposes.
"""

from __future__ import annotations


#: The version of RBIntegrations.
#:
#: This is in the format of:
#:
#: (Major, Minor, Micro, Patch, alpha/beta/rc/final, Release Number, Released)
#:
VERSION: tuple[int, int, int, int, str, int, bool] = \
    (5, 0, 1, 0, 'final', 0, True)


def get_version_string() -> str:
    """Return the version as a human-readable string.

    Returns:
        str:
        The human-readable version.
    """
    major, minor, micro, patch, tag, release_num, released = VERSION

    version = f'{major}.{minor}'

    if micro or patch:
        version += f'.{micro}'

    if patch:
        version += f'.{patch}'

    if tag != 'final':
        if tag == 'rc':
            version += f' RC{release_num}'
        else:
            version += f' {tag} {release_num}'

    if not released:
        version += ' (dev)'

    return version


def get_package_version() -> str:
    """Return the version as a Python package version string.

    Returns:
        str:
        The package version.
    """
    major, minor, micro, patch, tag, release_num = VERSION[:6]

    version = f'{major}.{minor}'

    if micro or patch:
        version += f'.{micro}'

    if patch:
        version += f'.{patch}'

    if tag != 'final':
        version += f'{tag}{release_num}'

    return version


def is_release() -> bool:
    """Return whether this is a released version.

    Returns:
        bool:
        ``True`` if this is a released version, or ``False`` if it's still
        in development.
    """
    return VERSION[6]


#: An alias for the the version information from :py:data:`VERSION`.
#:
#: This does not include the last entry in the tuple (the released state).
__version_info__ = VERSION[:-1]


#: An alias for the version used for the Python package.
__version__ = get_package_version()
